use std::sync::Arc;

use actix::Addr;
use actix_web::{web, HttpResponse, Responder};

use crate::actors::providers_states_actor;
use crate::actors::providers_states_actor::{ProviderLoginState, ProvidersStatesActor};
use crate::controllers::base_controller::{build_fatal_error_page, redirect_user};
use crate::data::action_logger::{Action, ActionLogger};
use crate::data::login_redirect::LoginRedirect;
use crate::data::provider::{ProviderID, ProvidersManager};
use crate::data::provider_configuration::ProviderConfigurationHelper;
use crate::data::remote_ip::RemoteIP;

#[derive(serde::Deserialize)]
pub struct StartLoginQuery {
    #[serde(default)]
    redirect: LoginRedirect,
    id: ProviderID,
}

/// Start user authentication using a provider
#[allow(clippy::too_many_arguments)]
pub async fn start_login(
    remote_ip: RemoteIP,
    providers: web::Data<Arc<ProvidersManager>>,
    states: web::Data<Addr<ProvidersStatesActor>>,
    query: web::Query<StartLoginQuery>,
    logger: ActionLogger,
) -> impl Responder {
    // Get provider information
    let provider = match providers.find_by_id(&query.id) {
        None => {
            return HttpResponse::NotFound()
                .body(build_fatal_error_page("Login provider not found!"));
        }
        Some(p) => p,
    };

    // Generate & save state
    let state = ProviderLoginState::new(&provider.id, query.redirect.clone());
    states
        .send(providers_states_actor::RecordState {
            ip: remote_ip.0,
            state: state.clone(),
        })
        .await
        .unwrap();

    logger.log(Action::StartLoginAttemptWithOpenIDProvider {
        provider_id: &provider.id,
        state: &state.state_id,
    });

    // Get provider configuration
    let config = match ProviderConfigurationHelper::get_configuration(&provider).await {
        Ok(c) => c,
        Err(e) => {
            log::error!("Failed to load provider configuration! {}", e);
            return HttpResponse::InternalServerError().body(build_fatal_error_page(
                "Failed to load provider configuration!",
            ));
        }
    };

    log::debug!("Provider configuration: {:?}", config);

    let url = config.auth_url(&provider, &state);
    log::debug!("Redirect user on {url} for authentication",);

    // Redirect user
    redirect_user(&url)
}
